"""Miscellaneous utility functions."""

import jax
import nonlinear_benchmarks
import numpy as np

from freq_statespace._data_manager import InputOutputData, create_data_object
from freq_statespace._model_structures import ModelBLA, ModelNonlinearLFR
from freq_statespace._solve import SolveResult


def load_and_preprocess_silverbox_data() -> InputOutputData:
    """Load (from `nonlinear_benchmarks`) and preprocesses the Silverbox dataset."""
    # Load data
    train = nonlinear_benchmarks.Silverbox()[0]
    u, y = train.u, train.y

    N = 8192  # number of samples per period
    R = 6  # number of random phase multisine realizations
    P = 1  # number of periods

    nu, ny = 1, 1  # SISO system

    fs = 1e7 / 2**14  # [Hz]
    f_idx = np.arange(1, 2 * 1342, 2)  # excited odd harmonics

    # Process data
    N_init = 164  # number of initial samples to be discarded
    N_z = 100  # number of zero samples separating the blocks visually
    N_tr = 400  # number of transient samples

    u_train = np.zeros((N, R))
    y_train = np.zeros((N, R))
    for k in range(R):
        if k == 0:
            u = u[N_init:]
            y = y[N_init:]
        else:
            idx = N_z + N_tr
            u = u[idx:]
            y = y[idx:]

        u_train[:, k] = u[:N]
        y_train[:, k] = y[:N]

        u = u[N:]
        y = y[N:]

    # Reshape data to required dimensions
    u_train = u_train.reshape(N, nu, R, P)
    y_train = y_train.reshape(N, ny, R, P)

    return create_data_object(u_train, y_train, f_idx, fs)


def evaluate_model_performance(
    model: ModelBLA | ModelNonlinearLFR,
    data: InputOutputData,
    *,
    solve_result: SolveResult | None = None,
    offset: int | None = None,
) -> None:
    """Simulate model and prints NRMSEs, along with solver timings (if provided).

    Parameters
    ----------
    model : `ModelBLA` or `ModelNonlinearLFR`
        Model to be evaluated. Can be a BLA or NL-LFR model.
    data : `InputOutputData`
        Measured input-output data.
    solve_result : `SolveResult`, optional
        Result of the model solving process, containing timing and loss
        information.
    offset : int, optional
        A non-negative integer `≤ N`. This value is used to select
        the unknown initial state of the time-domain simulations. Specifically,
        the initial state is selected from the steady-state BLA state
        simulations as `x0 = x_bla[-offset, :, :]`. The number of input samples
        are prepended accordingly.  If `offset` is not specified, two entire
        periods are simulated from zero initial state. The first period is
        then discarded when computing the NRMSE.

    """
    if not isinstance(model, ModelBLA | ModelNonlinearLFR):
        raise TypeError("`model` must be either `ModelBLA` or `ModelNonlinearLFR`.")

    u, y = data.time.u, data.time.y
    N, ny = y.shape[:2]

    # Determine offset and initial state
    if offset is None:
        offset = N
        x0 = None
    else:
        if isinstance(model, ModelNonlinearLFR):
            bla = super(ModelNonlinearLFR, model)
        else:
            bla = model

        # Obtain steady-state BLA state simulations
        x_bla = bla._simulate(u, offset=N)[1]
        x0 = x_bla[-offset, :, :]

    # Simulate model
    y_sim = model._simulate(u, offset=offset, x0=x0)[0]

    # Compute NRMSE per output channel
    error = y - y_sim
    mse = np.mean(error**2, axis=(0, 2))
    norm = np.mean(y**2, axis=(0, 2))
    nrmse = 100 * np.sqrt(mse / norm)  # as a percentage

    if solve_result is not None:
        avg_time = solve_result.iter_times.mean()
        unit = "s"
        if avg_time < 1:
            avg_time *= 1000
            unit = "ms"

        print(
            f"Optimization completed in {solve_result.wall_time:.2f}s "
            f"({solve_result.iter_count} iterations, "
            f"{avg_time:.2f}{unit}/iter).\n"
        )

    name = "NL-LFR" if isinstance(model, ModelNonlinearLFR) else "BLA"

    if ny == 1:
        print(f"{name} simulation error: {nrmse[0]:.2f}%.")
    else:
        print(f"{name} simulation error:")
        for k in range(ny):
            print(f"    output {k + 1}: {nrmse[k]:.2f}%.")
    print("")


def get_key(seed: int, tag: str) -> jax.Array:
    """Generate a deterministic key from a base seed and a tag."""
    tag = hash(tag) & 0xFFFFFFFF  # ensure it's in 32-bit range
    return jax.random.fold_in(jax.random.key(seed), tag)



"""NL-LFR model inference and learning, and nonlinear optimization."""

from typing import NamedTuple

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import optimistix as optx

from freq_statespace import _misc
from freq_statespace._config import SEED, SOLVER
from freq_statespace._data_manager import FrequencyData, InputOutputData
from freq_statespace._model_structures import ModelBLA, ModelNonlinearLFR
from freq_statespace._solve import solve
from freq_statespace.f_static._feature_maps import AbstractFeatureMap
from freq_statespace.f_static._nonlin_funcs import (
    AbstractNonlinearFunction,
    BasisFunctionModel,
)


MAX_ITER_INFERENCE_AND_LEARNING = 1000
MAX_ITER_OPTIMIZATION = 100
EPSILON = 1e-10


class _ThetaWZ(eqx.Module):
    """Decision variables for inference and learning."""

    B_w_star: jnp.ndarray = eqx.field(converter=jnp.asarray)
    C_z_star: jnp.ndarray = eqx.field(converter=jnp.asarray)
    D_yw_star: jnp.ndarray = eqx.field(converter=jnp.asarray)
    D_zu_star: jnp.ndarray = eqx.field(converter=jnp.asarray)


class _ArgsInferenceLearning(NamedTuple):
    """Static arguments for inference and learning."""

    theta_uy: tuple  # (A, B_u, C_y)
    phi: AbstractFeatureMap
    lambda_w: float
    fixed_point_iters: int
    f_data: tuple  # (frequencies, sampling frequency, U, Y, G_yu)
    Lambda: jnp.ndarray  # shape (F, ny, ny)
    Tz_inv: jnp.ndarray  # shape (nz, nz)
    epsilon: float
    N: int


class _ArgsNonlinOptimization(NamedTuple):
    """Static arguments for nonlinear optimization."""

    theta_static: ModelNonlinearLFR
    u: jnp.ndarray  # shape (N, nu, R)
    Y: jnp.ndarray  # shape (N, ny, R)
    Lambda: jnp.ndarray  # shape (F, ny, ny)
    x0: jnp.ndarray  # shape (nx, R)
    offset: int


def _compute_weighting_matrix(freq: FrequencyData) -> jnp.ndarray:
    """Compute weighting matrix for the loss function."""
    F, ny = freq.Y.shape[:2]
    var_noise = freq.Y_var_noise

    # Each (ny x ny) matrix is diagonal with elements equal to the inverse of the noise
    # variance for each output at each frequency. If only one is available (`P==1`),
    # identity matrices are returned instead.
    Lambda = np.zeros((F, ny, ny))
    for k in range(F):
        diag = np.eye(ny) if var_noise is None else np.diag(1 / var_noise[k])
        Lambda[k, ...] = diag

    return jnp.asarray(Lambda)


def _compute_bla_loss(bla: ModelBLA, data: InputOutputData) -> float:
    """Compute the BLA performance on the loss function that is used in this module."""
    Lambda = _compute_weighting_matrix(data.freq)
    Y = data.freq.Y
    U = data.freq.U

    Y_hat = bla._frequency_response(data.freq.f) @ U
    loss = _compute_weighted_residual(Y, Y_hat, Lambda)
    loss_scalar = jnp.sum(jnp.abs(loss) ** 2)

    return loss_scalar


def _create_basis_function_model_given_beta(
    nw: int, phi: AbstractFeatureMap, beta: jnp.ndarray
) -> BasisFunctionModel:
    """Create a `BasisFunctionModel` instance given `beta`.

    When a user does not want to perform inference and learning, and instead
    wants to perform nonlinear optimization directly, a `BasisFunctionModel` is
    initialized randomly with a seed (specifying a custom `beta` is not
    supported). This (private!) function bypasses that behavior by replacing
    the random `beta` with the provided matrix obtained from inference and
    learning. A dummy seed is still required for initialization, but it does
    not affect the final result.
    """
    dummy_seed = 0

    return eqx.tree_at(
        where=lambda tree: tree.beta,
        pytree=BasisFunctionModel(nw, phi, dummy_seed),
        replace=beta,
    )


def _prepare_inference_and_learning(
    bla: ModelBLA,
    data: InputOutputData,
    phi: AbstractFeatureMap,
    nw: int,
    lambda_w: float,
    fixed_point_iters: int,
    seed: int,
    epsilon: float,
) -> tuple[ModelNonlinearLFR, _ArgsInferenceLearning]:
    """Prepare initial guess and function arguments for inference and learning."""
    nz = phi.nz
    ny, nx = bla.C_y.shape
    N, nu = data.time.u.shape[:2]

    f_full = data.freq.f
    U = data.freq.U
    Y = data.freq.Y

    # Initialize theta_wz
    key = jax.random.key(seed)
    key_B_w, key_C_z, key_D_yw, key_D_zu = jax.random.split(key, 4)

    B_w_star = jax.random.normal(key_B_w, (nx, nw))
    C_z_star = jax.random.normal(key_C_z, (nz, nx))
    D_zu_star = jax.random.normal(key_D_zu, (nz, nu))
    D_yw_star = jax.random.normal(key_D_yw, (ny, nw))

    theta_wz = _ThetaWZ(B_w_star, C_z_star, D_yw_star, D_zu_star)
    theta_uy = (bla.A, bla.B_u, bla.C_y)

    # Compute z_star normalization
    G_zu = ModelBLA(  # note that this is not really the BLA
        A=bla.A, B_u=bla.B_u, C_y=C_z_star, D_yu=D_zu_star, ts=bla.ts, norm=bla.norm
    )._frequency_response(f_full)

    Z_star = G_zu @ U
    z_star = np.fft.irfft(Z_star, axis=0)
    z_star_min, z_star_max = z_star.min(axis=(0, 2)), z_star.max(axis=(0, 2))
    Tz_inv = jnp.diag(2 / (z_star_max - z_star_min))

    G_yu = bla._frequency_response(f_full)
    f_data = (f_full, 1 / data.time.ts, U, Y, G_yu)

    args = _ArgsInferenceLearning(
        theta_uy=theta_uy,
        phi=phi,
        lambda_w=lambda_w,
        fixed_point_iters=fixed_point_iters,
        f_data=f_data,
        Lambda=_compute_weighting_matrix(data.freq),
        Tz_inv=Tz_inv,
        epsilon=epsilon,
        N=N,
    )

    return theta_wz, args


def _prepare_nonlin_optimization(
    data: InputOutputData, model_init: ModelNonlinearLFR, offset: int
) -> tuple[ModelNonlinearLFR, _ArgsNonlinOptimization]:
    """Prepare initial guess and function arguments for nonlinear optimization."""
    theta0, theta_static = eqx.partition(model_init, eqx.is_inexact_array)

    bla = super(ModelNonlinearLFR, model_init)

    # Select the initial state from steady-state BLA simulations
    x_bla = bla._simulate(data.time.u, offset=data.time.u.shape[0])[1]
    x0 = x_bla[-offset, :, :]

    args = _ArgsNonlinOptimization(
        theta_static=theta_static,
        u=data.time.u,
        Y=data.freq.Y,
        Lambda=_compute_weighting_matrix(data.freq),
        x0=x0,
        offset=offset,
    )

    return theta0, args


def _compute_weighted_residual(
    Y: jnp.ndarray, Y_hat: jnp.ndarray, Lambda: jnp.ndarray
) -> jnp.ndarray:
    """Compute the weighted residuals between the measured and predicted outputs.

    Parameters
    ----------
    Y : jnp.ndarray, shape (N, ny, R)
        Measured output spectrum, averaged over periods.
    Y_hat : jnp.ndarray, shape (N, ny, R)
        Simulated output spectrum.
    Lambda : jnp.ndarray, shape (F, ny, ny)
        Weight matrix.

    """
    N, _, R = Y.shape
    return jnp.sqrt(Lambda / (R * N)) @ (Y - Y_hat)


def _loss_inference_and_learning(
    theta: _ThetaWZ, args: _ArgsInferenceLearning
) -> tuple:
    """Implement the inference and learning method and compute the loss."""
    f_full, fs, U, Y, G_yu = args.f_data

    A = args.theta_uy[0]
    B_u = args.theta_uy[1]
    C_y = args.theta_uy[2].astype(complex)

    B_w = theta.B_w_star.astype(complex)
    C_z = (args.Tz_inv @ theta.C_z_star).astype(complex)
    D_yw = theta.D_yw_star
    D_zu = args.Tz_inv @ theta.D_zu_star

    nw = D_yw.shape[1]
    nz, nu = D_zu.shape
    F = U.shape[0]
    R = U.shape[2]

    Theta = jnp.vstack((B_w, D_yw)).T @ jnp.vstack((B_w, D_yw))
    Theta += args.epsilon / args.lambda_w * jnp.eye(nw)

    z = 2 * jnp.pi * f_full / fs
    zj = jnp.exp(z * 1j)

    I_nx = jnp.eye(A.shape[0])

    def _compute_parametric_Gs(k):
        G_x = jnp.linalg.solve(zj[k] * I_nx - A, jnp.hstack((B_u, B_w)))
        return (
            C_y @ G_x[:, nu:] + D_yw,  # G_yw
            C_z @ G_x[:, :nu] + D_zu,  # G_zu
            C_z @ G_x[:, nu:],  # G_zw
        )

    G_yw, G_zu, G_zw = jax.vmap(_compute_parametric_Gs)(jnp.arange(F))

    # 1) Nonparametric inference
    def _infer_nonparametric_signals(k):
        Psi = G_yw[k, ...].T @ args.Lambda[k, ...]
        Phi = Psi @ G_yw[k, ...] + args.lambda_w * Theta
        W_hat = jnp.linalg.solve(Phi, Psi @ (Y[k, ...] - G_yu[k, ...] @ U[k, ...]))
        Z_hat = G_zu[k, ...] @ U[k, ...] + G_zw[k, ...] @ W_hat
        Y_hat = G_yu[k, ...] @ U[k, ...] + G_yw[k, ...] @ W_hat
        return W_hat, Z_hat, Y_hat

    W_star, Z_star, Y_hat = jax.vmap(_infer_nonparametric_signals)(jnp.arange(F))  # noqa: E501

    # 2) Parametric learning
    w_star = jnp.fft.irfft(W_star, n=args.N, axis=0)
    z_star = jnp.fft.irfft(Z_star, n=args.N, axis=0)

    w_star_stacked = jnp.transpose(w_star, (2, 0, 1)).reshape(args.N * R, nw)
    z_star_stacked = jnp.transpose(z_star, (2, 0, 1)).reshape(args.N * R, nz)

    # 2a) Compute beta
    phi_z_star = args.phi._compute_features(z_star_stacked)
    beta_hat = jnp.linalg.solve(
        phi_z_star.T @ phi_z_star, phi_z_star.T @ w_star_stacked
    )

    # 2b) Perform fixed-point iterations
    def _fixed_point_iteration(_, phi_z):
        w_stacked = phi_z @ beta_hat
        w = jnp.transpose(w_stacked.reshape(R, args.N, nw), (1, 2, 0))
        W = jnp.fft.rfft(w, axis=0)
        Z = G_zu @ U + G_zw @ W
        z = jnp.fft.irfft(Z, n=args.N, axis=0)
        z_stacked = jnp.transpose(z, (2, 0, 1)).reshape(args.N * R, nz)
        return args.phi._compute_features(z_stacked)

    phi_z = jax.lax.fori_loop(
        0,
        args.fixed_point_iters,
        _fixed_point_iteration,
        phi_z_star,
        unroll=True,  # noqa: E501
    )

    w_hat_stacked = phi_z @ beta_hat
    w_hat = jnp.transpose(w_hat_stacked.reshape(R, args.N, nw), (1, 2, 0))
    W_beta = jnp.fft.rfft(w_hat, axis=0)

    # 2c) Compute loss
    Y_hat = G_yu @ U + G_yw @ W_beta
    loss = _compute_weighted_residual(Y, Y_hat, args.Lambda)
    loss_real = (loss.real, loss.imag)
    scalar_loss = jnp.sum(jnp.abs(loss) ** 2)

    return loss_real, (scalar_loss, beta_hat)


def _loss_nonlin_optimization(
    theta: ModelNonlinearLFR, args: _ArgsNonlinOptimization
) -> tuple:
    """Perform time-domain forward simulations and compute the loss."""
    theta = eqx.combine(theta, args.theta_static)

    # Simulate the model in time domain
    y_hat = theta._simulate(args.u, offset=args.offset, x0=args.x0)[0]
    Y_hat = jnp.fft.rfft(y_hat, axis=0)

    # Compute loss
    loss = _compute_weighted_residual(args.Y, Y_hat, args.Lambda)
    loss_real = (loss.real, loss.imag)
    scalar_loss = jnp.sum(jnp.abs(loss) ** 2)

    return loss_real, (scalar_loss,)


def inference_and_learning(
    bla: ModelBLA,
    data: InputOutputData,
    *,
    phi: AbstractFeatureMap,
    nw: int,
    lambda_w: float,
    fixed_point_iters: int,
    solver: optx.AbstractLeastSquaresSolver | optx.AbstractMinimiser = SOLVER,
    max_iter: int = MAX_ITER_INFERENCE_AND_LEARNING,
    seed: int = SEED,
    epsilon: float = EPSILON,
) -> ModelNonlinearLFR:
    """Perform inference and learning.

    Parameters
    ----------
    bla : `ModelBLA`
        Parametric BLA model; is kept fixed during optimization.
    data : `InputOutputData`
        Measured input-output data object.
    phi : `AbstractFeatureMap`
        Any feature mapping that is linear in the parameters.
    nw : int
        Dimension of the latent signal `w`.
    lambda_w : float
        Regularization weight that controls the solution variance of `w`
        during inference.
    fixed_point_iters : int
        Number of fixed-point iterations for internal consistency.
    solver : `optx.AbstractLeastSquaresSolver` or `optx.AbstractMinimiser`
        Any least-squares solver or general minimization solver from the
        Optimistix or Optax libraries.
    max_iter : int
        Maximum number of optimization iterations.
    seed : int
        Random seed for parameter initialization.
    epsilon : float
        Numerical regularization constant for matrix inversion.

    Returns
    -------
    `ModelNonlinearLFR`
        Fully initialized NL-LFR model.

    """
    header = " NL-LFR inference and learning  "
    print(f"{header:=^72}")

    theta0, args = _prepare_inference_and_learning(
        bla, data, phi, nw, lambda_w, fixed_point_iters, seed, epsilon
    )
    bla_loss = _compute_bla_loss(bla, data)

    # Optimize the model parameters
    print("Starting iterative optimization...")
    print(f"    BLA loss: {bla_loss:.4e}")
    solve_result = solve(theta0, solver, args, _loss_inference_and_learning, max_iter)

    aux = solve_result.aux
    scalar_loss = aux[0]
    beta = aux[-1]

    model = ModelNonlinearLFR(
        A=args.theta_uy[0],
        B_u=args.theta_uy[1],
        C_y=args.theta_uy[2],
        D_yu=bla.D_yu,
        B_w=solve_result.theta.B_w_star,
        C_z=args.Tz_inv @ solve_result.theta.C_z_star,
        D_yw=solve_result.theta.D_yw_star,
        D_zu=args.Tz_inv @ solve_result.theta.D_zu_star,
        f_static=_create_basis_function_model_given_beta(nw, phi, beta),
        ts=data.time.ts,
        norm=bla.norm,
    )

    _misc.evaluate_model_performance(model, data, solve_result=solve_result)

    # Compute recursive loss of the NL-LFR model to check consistency with the
    # inference and learning solution. We simulate 2 periods and discard the
    # first one to ensure that transients have died out (see `offset`)
    y_hat = model._simulate(data.time.u, offset=data.time.u.shape[0])[0]
    Y_hat = jnp.fft.rfft(y_hat, axis=0)
    loss_recursive = _compute_weighted_residual(data.freq.Y, Y_hat, args.Lambda)
    scalar_loss_recursive = jnp.sum(jnp.abs(loss_recursive) ** 2)

    print("NL-LFR loss consistency check:")
    print(f"    Inference & learning loss = {scalar_loss:.4e}")
    print(f"    Recursive simulation loss = {scalar_loss_recursive:.4e}\n")

    return model


def optimize(
    model: ModelNonlinearLFR,
    data: InputOutputData,
    *,
    solver: optx.AbstractLeastSquaresSolver | optx.AbstractMinimiser = SOLVER,
    max_iter: int = MAX_ITER_OPTIMIZATION,
    offset: int | None = None,
) -> ModelNonlinearLFR:
    """Refine the parameters of an NL-LFR model using time-domain simulations.

    Parameters
    ----------
    model : `ModelNonlinearLFR`
        Initial NL-LFR model to optimize.
    data : `InputOutputData`
        Measured input-output data object.
    solver : `optx.AbstractLeastSquaresSolver` or `optx.AbstractMinimiser`
        Any least-squares solver or general minimization solver from the
        Optimistix or Optax libraries.
    max_iter : int
        Maximum number of optimization iterations.
    offset : int, optional
        A non-negative integer `≤ N`. This value is used to select
        the unknown initial state of the time-domain simulations. Specifically,
        the initial state is selected from the steady-state BLA state
        simulations as `x0 = x_bla[-offset, :, :]`. The number of input samples
        are prepended accordingly. This warm-up ensures the system reaches
        steady-state before the main period begins, allowing for leakage-free
        DFT computations (the prepended samples are discarded). This approach
        is valid because the data is assumed to be periodic. If not specified,
        `offset` defaults to 10% of the data length.

    Returns
    -------
    `ModelNonlinearLFR`
        NL-LFR model with optimized parameters.

    """
    header = " NL-LFR optimization  "
    print(f"{header:=^72}")

    if offset is None:  # we start 10% 'ahead of time'
        offset = int(np.ceil(0.1 * data.time.u.shape[0]))

    theta0, args = _prepare_nonlin_optimization(data, model, offset)

    bla_loss = _compute_bla_loss(super(ModelNonlinearLFR, model), data)

    print("Starting iterative optimization...")
    print(f"    BLA loss: {bla_loss:.4e}")
    solve_result = solve(theta0, solver, args, _loss_nonlin_optimization, max_iter)

    model = eqx.combine(solve_result.theta, args.theta_static)

    _misc.evaluate_model_performance(
        model, data, solve_result=solve_result, offset=offset
    )

    return model


def construct(bla: ModelBLA, f_static: AbstractNonlinearFunction) -> ModelNonlinearLFR:
    """Construct an NL-LFR model given the BLA and a static nonlinear function object.

    Parameters
    ----------
    bla : `ModelBLA`
        Parametric BLA model.
    f_static : `AbstractNonlinearFunction`
        Nonlinear function object that defines the static relation between
        the latent signals `z` and `w`. Note that the `seed` attribute of
        `f_static` is used to randomly initialize the matrices `B_w`, `C_z`,
        `D_yw`, and `D_zu`.

    Returns
    -------
    `ModelNonlinearLFR`
        NL-LFR model with (partly random) initial parameters.

    """
    nw, nz = f_static.nw, f_static.nz
    ny, nu = bla.D_yu.shape
    nx = bla.A.shape[0]

    key = _misc.get_key(f_static.seed, "nonlin_lfr")

    key_B_w, key_C_z, key_D_yw, key_D_zu = jax.random.split(key, 4)

    # We want our initial loss to be very close to the BLA loss. To achieve
    # this, we make the elements of B_w and D_yw small enough so the
    # contribution of the static nonlinearity does not significantly affect
    # the initial loss. (Initializing B_w and D_yw with zeros could hamper
    # convergence because of zero gradients.)
    scale = 1e-6
    B_w = scale * jax.random.normal(key_B_w, (nx, nw))
    D_yw = scale * jax.random.normal(key_D_yw, (ny, nw))

    C_z = jax.random.normal(key_C_z, (nz, nx))
    D_zu = jax.random.normal(key_D_zu, (nz, nu))

    return ModelNonlinearLFR(
        bla.A,
        bla.B_u,
        C_y=bla.C_y,
        D_yu=bla.D_yu,
        B_w=B_w,
        C_z=C_z,
        D_yw=D_yw,
        D_zu=D_zu,
        f_static=f_static,
        ts=bla.ts,
        norm=bla.norm,
    )



"""General-purpose wrapper/logger around Optimistix solvers."""
import time
from dataclasses import dataclass
from typing import Any, cast

import equinox as eqx
import jax
import numpy as np
import optimistix as optx
from jaxtyping import PyTree, Scalar
from optimistix._custom_types import Aux, Fn, Out, Y
from optimistix._least_squares import _ToMinimiseFn
from optimistix._misc import OutAsArray


@dataclass(frozen=True)
class SolveResult:
    """Result of the optimization process.

    Attributes
    ----------
    theta : Any
        Final solution vector or PyTree of parameters.
    aux : tuple
        Auxiliary output from the final loss evaluation. The first element is
        always the scalar loss value.
    loss_history : np.ndarray
        Sequence of loss values at each iteration.
    iter_count : int
        Number of iterations executed before termination.
    iter_times : np.ndarray
        Wall-clock duration (in seconds) for each iteration.
    converged : bool
        Whether the termination criterion was met.
    wall_time : float
        Total wall-clock time (in seconds) spent in the optimization loop.

    """

    theta: Y
    aux: Any
    loss_history: np.ndarray
    iter_count: int
    iter_times: np.ndarray
    converged: bool
    wall_time: float


def solve(
    theta_init: Y,
    solver: optx.AbstractLeastSquaresSolver | optx.AbstractMinimiser,
    args: PyTree[Any],
    loss_fn: Fn,
    max_iter: int
) -> SolveResult:
    """Solve an optimization problem using a JAX-compatible Optimistix solver.

    Supports both least-squares and general minimization solvers. Tracks
    iteration timing and loss history. Applies JIT compilation to improve
    performance.

    Parameters
    ----------
    theta_init : Any
        Initial guess for the parameters to optimize (PyTree).
    solver : `AbstractLeastSquaresSolver` or `AbstractMinimiser`
        Solver instance from the Optimistix library.
    args : PyTree[Any]
        Additional static/dynamic arguments passed to the loss function.
    loss_fn : Callable
        Objective function to be minimized. Must return a tuple: (loss, aux).
    max_iter : int
        Maximum number of iterations allowed before termination.

    Returns
    -------
    SolveResult
        Structured result containing final parameters, loss trace, timing,
        and convergence information.

    """
    loss_fn = OutAsArray(loss_fn)
    if isinstance(solver, optx.AbstractMinimiser):
        loss_fn = _ToMinimiseFn(loss_fn)
        loss_fn = eqx.filter_closure_convert(loss_fn, theta_init, args)
        loss_fn = cast(Fn[Y, Scalar, Aux], loss_fn)
    elif isinstance(solver, optx.AbstractLeastSquaresSolver):
        loss_fn = eqx.filter_closure_convert(loss_fn, theta_init, args)
        loss_fn = cast(Fn[Y, Out, Aux], loss_fn)
    else:
        raise ValueError('Unknown solver type.')

    tags = frozenset()
    f_struct, aux_struct = loss_fn.out_struct
    options = {}

    # JIT compile step and terminate
    step = eqx.filter_jit(
        eqx.Partial(solver.step, fn=loss_fn, args=args, options=options, tags=tags)  # noqa: E501
    )
    terminate = eqx.filter_jit(
        eqx.Partial(solver.terminate, fn=loss_fn, args=args, options=options, tags=tags)  # noqa: E501
    )

    # Initial state
    state = solver.init(loss_fn, theta_init, args, options, f_struct, aux_struct, tags)  # noqa: E501
    converged = terminate(y=theta_init, state=state)[0]

    iter_count = 0
    theta = theta_init
    loss_history = np.zeros((max_iter,))
    iter_times = np.zeros((max_iter,))

    # Warm up JIT compilation
    _ = step(y=theta, state=state)

    start_time = time.time()

    while not converged and iter_count < max_iter:
        iter_start = time.perf_counter()
        theta, state, aux = step(y=theta, state=state)
        iter_end = time.perf_counter()

        scalar_loss = aux[0]
        jax.debug.print("    Iter {0} | loss = {1:.4e}", iter_count + 1, scalar_loss)  # noqa: E501

        loss_history[iter_count] = scalar_loss
        iter_times[iter_count] = iter_end - iter_start

        converged = terminate(y=theta, state=state)[0]
        iter_count += 1

    wall_time = time.time() - start_time

    return SolveResult(theta, aux, loss_history[:iter_count], iter_count,
                       iter_times[:iter_count], converged, wall_time)



"""BLA and NL-LFR model classes, optimized for use with JAX and Equinox."""

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np

from freq_statespace._data_manager import Normalizer
from freq_statespace.f_static._nonlin_funcs import AbstractNonlinearFunction


class ModelBLA(eqx.Module):
    """BLA model class.

    Attributes
    ----------
    A : jnp.ndarray, shape (nx, nx)
    B_u : jnp.ndarray, shape (nx, nu)
    C_y : jnp.ndarray, shape (ny, nx)
    D_yu : jnp.ndarray, shape (ny, nu)
    ts : float
        Sampling time (in seconds) of the discrete system.
    norm : Normalizer
        Contains means and standard deviations of input-output signals.

    """

    A: jnp.ndarray = eqx.field(converter=jnp.asarray)
    B_u: jnp.ndarray = eqx.field(converter=jnp.asarray)
    C_y: jnp.ndarray = eqx.field(converter=jnp.asarray)
    D_yu: jnp.ndarray = eqx.field(converter=jnp.asarray)
    ts: float
    norm: Normalizer = eqx.field(static=True)

    def _simulate(
        self, u: np.ndarray, *, offset: int = 0, x0: np.ndarray | None = None
    ) -> tuple[jnp.ndarray, jnp.ndarray]:
        """Simulate the BLA model in the time domain.

        To be used within an optimization loop, as it assumes normalized data.

        Parameters
        ----------
        u : jnp.ndarray, shape (N, nu, R)
            Normalized input signal where:
            - N: number of time steps
            - nu: number of inputs
            - R: number of realizations
        offset : int
            Non-negative integer, must be `≤ N`. If greater than 0, the last
            `offset` input samples are prepended before the simulation starts.
            These prepended samples are used for simulation warm-up and are not
            included in the returned tuple.
        x0 : jnp.ndarray of shape (nx, R), optional
            Initial state of the system. If not provided, the simulation starts
            from a zero state. If `offset > 0`, the simulation first prepends
            the `offset` input samples; `x0` then refers to the initial state
            of the offset samples.

        Returns
        -------
        Y : jnp.ndarray, shape (N, ny, R)
            Simulated output trajectories.
        X : jnp.ndarray, of shape (N, nx, R)
            Simulated state trajectories.
        W : jnp.ndarray, of shape (N, nw, R)
            Static nonlinear function outputs.
        Z : jnp.ndarray, of shape (N, nz, R)
            Static nonlinear function inputs.

        Raises
        ------
        ValueError
            If `offset` is not a non-negative integer `≤ N`.

        """
        def _make_step(k, state):
            X, Y_accum, X_accum = state
            U = jax.lax.dynamic_slice(u, (k, 0, 0), (1, nu, R)).squeeze(axis=0)

            # Model equations
            X_next = self.A @ X + self.B_u @ U
            Y = self.C_y @ X + self.D_yu @ U
            return X_next, Y_accum.at[k, ...].set(Y), X_accum.at[k, ...].set(X)

        # Validate offset
        if not isinstance(offset, int) or offset < 0 or offset > u.shape[0]:
            raise ValueError(
                f"'offset' must be a non-negative integer smaller"
                f" or equal to N (={u.shape[0]}), got {offset}."
            )

        # Extend input signal if needed
        if offset > 0:
            u = jnp.concatenate((u[-offset:, ...], u), axis=0)

        N, nu, R = u.shape
        ny, nx = self.C_y.shape

        if x0 is None:
            x0 = jnp.zeros((nx, R))

        loop_init = (
            x0,
            jnp.zeros((N, ny, R)),  # Y_accum
            jnp.zeros((N, nx, R)),  # X_accum
        )
        Y, X = jax.lax.fori_loop(0, N, _make_step, loop_init)[1:]
        return Y[offset:, ...], X[offset:, ...]

    def simulate(
        self,
        u: np.ndarray,
        *,
        x0: np.ndarray | None = None,
    ) -> tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
        """Simulate the BLA model in the time domain.

        Parameters
        ----------
        u : np.ndarray, shape (N, nu) or (N,)
            Input signal of shape (N, nu). If the system has only one input
            channel, `u` may be a 1D array of shape (N,).
        x0 : np.ndarray, shape (nx,), optional
            Initial state of the system. If not provided, the initial state is
            assumed to be zero.

        Returns
        -------
        y : np.ndarray, shape (N, ny) or (N,)
            Simulated output trajectories. Shape is (N,) in case `ny == 1`.
        t : np.ndarray, shape (N,)
            Time vector corresponding to the simulation.
        x : np.ndarray, shape (N, nx) or (N,)
            Simulated state trajectories. Shape is (N,) in case `nx == 1`.

        """
        # Validate if `u` is 1D or 2D, then extend to 3D
        if u.ndim == 1:
            u = u[:, None, None]
        elif u.ndim == 2:
            u = u[:, :, None]
        else:
            raise ValueError(f"`u` must be 1D or 2D, got {u.ndim}D.")

        # Validate initial state
        if x0 is not None:
            if x0.ndim != 1:
                raise ValueError(f"`x0` must be 1D, got {x0.ndim}D.")
            if x0.shape[0] != self.A.shape[0]:
                raise ValueError(
                    f"`x0` must have shape ({self.A.shape[0]},), got {x0.shape}."
                )
            x0 = x0.reshape(-1, 1)
        else:
            x0 = jnp.zeros((self.A.shape[0], 1))

        # Normalize input signal
        u_mean = self.norm.u_mean.reshape(1, -1)
        u_std = self.norm.u_std.reshape(1, -1)
        u = (u - u_mean) / u_std

        u = jnp.asarray(u)
        y, x = self._simulate(u, offset=0, x0=x0)

        # Denormalize output signal
        y_mean = self.norm.y_mean.reshape(1, -1, 1)
        y_std = self.norm.y_std.reshape(1, -1, 1)
        y = y * y_std + y_mean

        y, x = jnp.squeeze(y), jnp.squeeze(x)

        t = np.arange(y.shape[0]) * self.ts

        return np.asarray(y), t, np.asarray(x)

    def _frequency_response(
        self,
        f: np.ndarray,
    ) -> jnp.ndarray:
        """Compute the frequency response of the system.

        To be used within an optimization loop, as it assumes normalized data.

        Parameters
        ----------
        f : np.ndarray, shape (F_f,)
            Frequency points in Hz.

        Returns
        -------
        G : jnp.ndarray
            Frequency response matrix of shape (F_f, ny, nu).

        """
        def G(k):
            G_x = jnp.linalg.solve(zj[k] * I_nx - self.A, B_u)
            return C_y @ G_x + self.D_yu

        fs = 1 / self.ts
        z = 2 * jnp.pi * f / fs
        zj = jnp.exp(z * 1j)

        I_nx = jnp.eye(self.A.shape[0])
        B_u = self.B_u.astype(complex)  # to suppress a warning
        C_y = self.C_y.astype(complex)  # to suppress a warning
        return jax.vmap(G)(np.arange(len(f)))

    def num_parameters(self) -> int:
        """Return the total number of model parameters."""
        return self.A.size + self.B_u.size + self.C_y.size + self.D_yu.size


class ModelNonlinearLFR(ModelBLA):
    """NL-LFR model class.

    Inherits from `ModelBLA` and adds linear matrices `B_w`, `C_z`, `D_yw`, `D_zu`, and
    static nonlinear feedback.

    Attributes
    ----------
    B_w : jnp.ndarray, shape (nx, nw)
    C_z : jnp.ndarray, shape (nz, nx)
    D_yw : jnp.ndarray, shape (ny, nw)
    D_zu : jnp.ndarray, shape (nz, nu)
    f_static : `AbstractNonlinearFunction`
        Static nonlinear function mapping `z` to `w`.

    """

    B_w: jnp.ndarray = eqx.field(converter=jnp.asarray)
    C_z: jnp.ndarray = eqx.field(converter=jnp.asarray)
    D_yw: jnp.ndarray = eqx.field(converter=jnp.asarray)
    D_zu: jnp.ndarray = eqx.field(converter=jnp.asarray)
    f_static: AbstractNonlinearFunction

    def _simulate(
        self, u: np.ndarray, *, offset: int = 0, x0: np.ndarray | None = None
    ) -> tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
        """Simulate the NL-LFR model in the time domain.

        To be used within an optimization loop, as it assumes normalized data.

        Parameters
        ----------
        u : jnp.ndarray, shape (N, nu, R)
            Normalized input signal where:
            - N: number of time steps
            - nu: number of inputs
            - R: number of realizations
        offset : int
            Non-negative integer, must be `≤ N`. If greater than 0, the last
            `offset` input samples are prepended before the simulation starts.
            These prepended samples are used for simulation warm-up and are not
            included in the returned tuple.
        x0 : jnp.ndarray of shape (nx, R), optional
            Initial state of the system. If not provided, the simulation starts
            from a zero state. If `offset > 0`, the simulation first prepends
            the `offset` input samples; `x0` then refers to the initial state
            of the offset samples.

        Returns
        -------
        Y : jnp.ndarray, shape (N, ny, R)
            Simulated output trajectories.
        X : jnp.ndarray, of shape (N, nx, R)
            Simulated state trajectories.
        W : jnp.ndarray, of shape (N, nw, R)
            Static nonlinear function outputs.
        Z : jnp.ndarray, of shape (N, nz, R)
            Static nonlinear function inputs.

        Raises
        ------
        ValueError
            If `offset` is not a non-negative integer `≤ N`.

        """
        def _make_step(k, state):
            X, Y_accum, X_accum, W_accum, Z_accum = state
            U = jax.lax.dynamic_slice(u, (k, 0, 0), (1, nu, R)).squeeze(axis=0)

            # Model equations
            Z = self.C_z @ X + self.D_zu @ U
            W = self.f_static._evaluate(Z.T).T
            X_next = self.A @ X + self.B_u @ U + self.B_w @ W
            Y = self.C_y @ X + self.D_yu @ U + self.D_yw @ W
            return (
                X_next,
                Y_accum.at[k, ...].set(Y),
                X_accum.at[k, ...].set(X),
                W_accum.at[k, ...].set(W),
                Z_accum.at[k, ...].set(Z),
            )

        # Validate offset
        if not isinstance(offset, int) or offset < 0 or offset > u.shape[0]:
            raise ValueError(
                f"'offset' must be a non-negative integer smaller"
                f" or equal to N (={u.shape[0]}), got {offset}."
            )

        # Extend input signal if needed
        if offset > 0:
            u = jnp.concatenate((u[-offset:, ...], u), axis=0)

        N, nu, R = u.shape
        nz, nx = self.C_z.shape
        ny, nw = self.D_yw.shape

        if x0 is None:
            x0 = jnp.zeros((nx, R))

        loop_init = (
            x0,
            jnp.zeros((N, ny, R)),  # Y_accum
            jnp.zeros((N, nx, R)),  # X_accum
            jnp.zeros((N, nw, R)),  # W_accum
            jnp.zeros((N, nz, R)),  # Z_accum
        )
        Y, X, W, Z = jax.lax.fori_loop(0, N, _make_step, loop_init)[1:]
        return (Y[offset:, ...], X[offset:, ...], W[offset:, ...], Z[offset:, ...])

    def simulate(
        self,
        u: np.ndarray,
        *,
        x0: np.ndarray | None = None,
    ) -> tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
        """Simulate the NL-LFR model in the time domain.

        Parameters
        ----------
        u : np.ndarray, shape (N, nu) or (N,)
            Input signal of shape (N, nu). If the system has only one input
            channel, `u` may be a 1D array of shape (N,).
        x0 : np.ndarray, shape (nx,), optional
            Initial state of the system. If not provided, the initial state is
            assumed to be zero.

        Returns
        -------
        y : np.ndarray, shape (N, ny) or (N,)
            Simulated output trajectories. Shape is (N,) in case `ny == 1`.
        t : np.ndarray, shape (N,)
            Time vector corresponding to the simulation.
        x : np.ndarray, shape (N, nx) or (N,)
            Simulated state trajectories. Shape is (N,) in case `nx == 1`.
        w : np.ndarray, shape (N, nw), or (N,)
            Static nonlinear function outputs. Shape is (N,) in case `nw == 1`.
        z : np.ndarray, shape (N, nz) or (N,)
            Static nonlinear function inputs. Shape is (N,) in case `nz == 1`.

        """
        # Validate if `u` is 1D or 2D, then extend to 3D
        if u.ndim == 1:
            u = u[:, None, None]
        elif u.ndim == 2:
            u = u[:, :, None]
        else:
            raise ValueError(f"`u` must be 1D or 2D, got {u.ndim}D.")

        # Validate initial state
        if x0 is not None:
            if x0.ndim != 1:
                raise ValueError(f"`x0` must be 1D, got {x0.ndim}D.")
            if x0.shape[0] != self.A.shape[0]:
                raise ValueError(
                    f"`x0` must have shape ({self.A.shape[0]},), got {x0.shape}."
                )
            x0 = x0.reshape(-1, 1)
        else:
            x0 = jnp.zeros((self.A.shape[0], 1))

        # Normalize input signal
        u_mean = self.norm.u_mean.reshape(1, -1)
        u_std = self.norm.u_std.reshape(1, -1)
        u = (u - u_mean) / u_std

        y, x, w, z = self._simulate(u, offset=0, x0=x0)

        # Denormalize output signal
        y_mean = self.norm.y_mean.reshape(1, -1, 1)
        y_std = self.norm.y_std.reshape(1, -1, 1)
        y = y * y_std + y_mean

        y, x = jnp.squeeze(y), jnp.squeeze(x)
        w, z = jnp.squeeze(w), jnp.squeeze(z)

        t = np.arange(y.shape[0]) * self.ts

        return np.asarray(y), t, np.asarray(x), np.asarray(w), np.asarray(z)

    def num_parameters(self) -> int:
        """Return the total number of model parameters."""
        return (
            self.B_w.size
            + self.C_z.size
            + self.D_yw.size
            + self.D_zu.size
            + super().num_parameters()
            + self.f_static.num_parameters()
        )



"""Top-level `freq_statespace` package.

This module exposes the main model structures, static nonlinear building blocks,
data utilities, and example preprocessing functions for frequency-domain system
identification.

Available submodules and components:
- `create_data_object`: Utility function to initialize the identification process by
   constructing an `InputOutputData` object.
- `lin`: Tools for creating and optimizing linear BLA models.
- `nonlin`: Tools for creating and optimizing nonlinear LFR models;
- `f_static`: Static nonlinear function approximators and feature maps;
- `ModelBLA`, `ModelNonlinearLFR`: Primary model structure classes;
- `load_and_preprocess_silverbox_data`: Example dataset loading and preprocessing.

"""


from . import _best_linear_approximation as lin
from . import _nonlin_lfr as nonlin
from . import f_static
from ._data_manager import create_data_object
from ._misc import load_and_preprocess_silverbox_data
from ._model_structures import ModelBLA, ModelNonlinearLFR


__all__ = [
    "lin",
    "nonlin",
    "f_static",
    "load_and_preprocess_silverbox_data",
    "create_data_object",
    "ModelBLA",
    "ModelNonlinearLFR"
]



"""Nonlinear feature mappings (`z` to `features`) that are linear in the parameters."""

from abc import abstractmethod
from itertools import combinations_with_replacement

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np


class AbstractFeatureMap(eqx.Module, strict=True):
    """Abstract base class for feature mappings.

    Subclasses must provide the attribute `nz`, and must implement the methods
    `_compute_features` and `num_features`.
    """

    nz: eqx.AbstractVar[int]

    @abstractmethod
    def _compute_features(self, z: jnp.ndarray) -> jnp.ndarray:
        """From size (-1, `nz`) to size (-1, `num_features()`)."""
        pass

    @abstractmethod
    def num_features(self) -> int:
        """Return the total number of model features."""
        pass


class Polynomial(AbstractFeatureMap, strict=True):
    """Flexible polynomial feature map.

    Attributes
    ----------
    nz : int
        Number of input features (dimension of latent signal `z`).
    degree : int
        Maximum polynomial degree.
    type : str
        Type of polynomial (`"full"`, `"odd"`, or `"even"`).
    cross_terms : bool
        If `True`, includes cross-terms in the polynomial features.
    offset : bool
        If `True`, includes a constant offset term in the features.
    linear : bool
        If `True`, includes linear terms in the polynomial features.
    tanh_clip : bool
        If `True`, applies tanh clipping to the input features.

    """

    nz: int
    degree: int
    type: str = "full"
    cross_terms: bool = True
    offset: bool = True
    linear: bool = True
    tanh_clip: bool = True

    # Post-init attributes
    _num_features: int = eqx.field(init=False, repr=False)
    _combination_matrix: jnp.ndarray = eqx.field(init=False, repr=False)

    def __post_init__(self):
        """Construct the feature architecture based on the specified parameters."""
        if self.type == 'full':
            active_degrees = range(1 if self.linear else 2, self.degree + 1)
        elif self.type == 'odd':
            active_degrees = range(1 if self.linear else 3, self.degree + 1, 2)
        elif self.type == 'even':
            active_degrees = range(2, self.degree + 1, 2)
        else:
            raise ValueError(
                'Invalid polynomial `type`. Must be "full", "odd", or "even".'
            )

        max_degree = active_degrees[-1]
        combination_matrix = []
        combination_row = np.full((max_degree,), -1, dtype=int)
        num_features = 1 if self.offset else 0
        for _, deg in enumerate(active_degrees):
            if self.cross_terms:
                combinations = tuple(
                    combinations_with_replacement(
                        range(self.nz), deg
                    )
                )
            else:
                combinations = tuple(
                    (i,) * deg for i in range(self.nz)
                )

            num_features += len(combinations)
            for _, comb in enumerate(combinations):
                combination_row[:deg] = comb
                combination_matrix.append(combination_row.copy())

        self._num_features = num_features
        self._combination_matrix = jnp.array(combination_matrix)

    def _compute_features(self, z: jnp.ndarray) -> jnp.ndarray:

        N, nz = z.shape
        if nz != self.nz:
            raise ValueError(
                'Input size does not match the basis function size: '
                '`z.shape[1] != nz`.'
            )

        if self.tanh_clip:
            z = jnp.tanh(z)

        # Augment input to make it consistent with the combination matrix
        z_augmented = jnp.hstack((z, jnp.ones((N, 1))))

        def _compute_phi_z(combination_idx):
            combination = self._combination_matrix[combination_idx]
            z_selected = jnp.take(z_augmented, combination.ravel(), axis=1)
            return jnp.prod(z_selected, axis=1)

        num_combs = self._combination_matrix.shape[0]
        phi_z = jax.vmap(_compute_phi_z, out_axes=1)(jnp.arange(num_combs))

        return jnp.hstack((jnp.ones((N, 1)), phi_z)) if self.offset else phi_z

    def num_features(self) -> int:
        """Return the total number of model features."""
        return self._num_features


class LegendrePolynomial(AbstractFeatureMap, strict=True):
    """Legendre polynomial.

    Orthogonal over the interval [-1, 1] with unit weighting in the univariate case.

    Attributes
    ----------
    nz : int
        Number of input features (dimension of latent signal `z`).
    degree : int
        Maximum polynomial degree.
    offset : bool
        If `True`, includes a constant offset term in the features.
    tanh_clip : bool
        If `True`, applies tanh clipping to the input features.

    """

    nz: int
    degree: int
    offset: bool = True
    tanh_clip: bool = True

    # Post-init attribute
    _num_features: int = eqx.field(init=False, repr=False)

    def __post_init__(self):
        """Construct the feature architecture based on the specified parameters."""
        self._num_features = (
            self.nz * self.degree + (1 if self.offset else 0)
        )

    def _compute_features(self, z: jnp.ndarray) -> jnp.ndarray:

        def _compute_phi_z(k, state):
            phi_z, phi_z_previous, phi_z_two_before = state
            phi_z_current = (
                (2 * k - 1) / k * z * phi_z_previous
                - (k - 1) / k * phi_z_two_before
            )
            phi_z = phi_z.at[k-2, ...].set(phi_z_current)
            return phi_z, phi_z_current, phi_z_previous

        if self.tanh_clip:
            z = jnp.tanh(z)

        N = z.shape[0]
        phi_z0 = jnp.zeros((self.degree - 1, N, self.nz))
        loop_init = (phi_z0, z, jnp.ones_like(z))

        phi_z = jax.lax.fori_loop(
            2, self.degree + 1, _compute_phi_z, loop_init, unroll=True
        )[0]
        phi_z = jnp.transpose(phi_z, (1, 2, 0)).reshape(N, -1)
        phi_z = jnp.hstack((jnp.ones((N, 1)), z, phi_z))
        return phi_z if self.offset else phi_z[:, 1:]

    def num_features(self) -> int:
        """Return the total number of model features."""
        return self._num_features


class ChebyshevPolynomial(AbstractFeatureMap, strict=True):
    """Chebyshev polynomial.

    Orthogonal over the interval [-1, 1] with respect to a weight function that depends
    on the polynomial `type`.

    Attributes
    ----------
    nz : int
        Number of input features (dimension of latent signal `z`).
    degree : int
        Maximum polynomial degree.
    type : int
        Type of Chebyshev polynomial:
        - `1`: First kind (orthogonal w.r.t. 1 / sqrt(1 - x²))
        - `2`: Second kind (orthogonal w.r.t. sqrt(1 - x²))
    offset : bool
        If `True`, includes a constant offset term in the features.
    tanh_clip : bool
        If `True`, applies tanh clipping to the input features.

    """

    nz: int
    degree: int
    type: int
    offset: bool = True
    tanh_clip: bool = True

    # Post-init attribute
    _num_features: int = eqx.field(init=False, repr=False)

    def __post_init__(self):
        """Construct the feature architecture based on the specified parameters."""
        if self.type not in [1, 2]:
            raise ValueError('Invalid polynomial `type`. Must be `1` or `2`.')
        self._num_features = (
            self.nz * self.degree + (1 if self.offset else 0)
        )

    def _compute_features(self, z: jnp.ndarray) -> jnp.ndarray:

        def _compute_phi_z(k, state):
            phi_z, phi_z_previous, phi_z_two_before = state
            phi_z_current = 2 * z * phi_z_previous - phi_z_two_before
            phi_z = phi_z.at[k-2, ...].set(phi_z_current)
            return phi_z, phi_z_current, phi_z_previous

        if self.tanh_clip:
            z = jnp.tanh(z)

        N = z.shape[0]
        phi_z0 = jnp.zeros((self.degree - 1, N, self.nz))
        loop_init = (phi_z0, self.type * z, jnp.ones_like(z))

        phi_z = jax.lax.fori_loop(
            2, self.degree + 1, _compute_phi_z, loop_init, unroll=True
        )[0]
        phi_z = jnp.transpose(phi_z, (1, 2, 0)).reshape(N, -1)
        phi_z = jnp.hstack((jnp.ones((N, 1)), self.type * z, phi_z))
        return phi_z if self.offset else phi_z[:, 1:]

    def num_features(self) -> int:
        """Return the total number of model features."""
        return self._num_features



"""Nonparametric BLA, parametric subspace identification, and optimizer."""

import equinox as eqx
import jax.numpy as jnp
import numpy as np
import optimistix as optx

from freq_statespace import _misc
from freq_statespace._config import SOLVER
from freq_statespace._data_manager import (
    FrequencyData,
    InputOutputData,
    NonparametricBLA,
)
from freq_statespace._frequency_response import compute_frequency_response
from freq_statespace._model_structures import ModelBLA
from freq_statespace._solve import solve
from freq_statespace.dep import fsid


MAX_ITER = 1000


def _compute_weighted_residual(theta_dyn: ModelBLA, args: tuple) -> tuple:
    """Compute the weighted residuals between the nonparametric and parametric BLA."""
    theta_static, G_nonpar, f_data, W = args

    theta = eqx.combine(theta_dyn, theta_static)

    G_par = theta._frequency_response(f_data)
    loss = jnp.sqrt(W / G_nonpar.size) * (G_par - G_nonpar)
    loss_real = (loss.real, loss.imag)
    scalar_loss = jnp.sum(jnp.abs(loss) ** 2)

    return loss_real, (scalar_loss,)


def _normalize_states(model: ModelBLA, freq: FrequencyData) -> ModelBLA:
    """Normalize BLA model states to have unit variance."""
    nx, nu = model.B_u.shape

    f_data = freq.f

    G_xu = ModelBLA(  # note that this is not really the BLA
        A=model.A,
        B_u=model.B_u,  # usual state dynamics
        C_y=np.eye(nx),
        D_yu=np.zeros((nx, nu)),  # full-state output
        ts=model.ts,
        norm=model.norm,
    )._frequency_response(f_data)

    X = G_xu @ freq.U
    x = np.fft.irfft(X, axis=0)
    x_std = np.std(x, axis=(0, 2))

    Tx = np.diag(x_std)
    Tx_inv = np.diag(1 / x_std)

    # Apply similarity transformation: x_norm = Tx_inv * x
    return ModelBLA(
        A=Tx_inv @ model.A @ Tx,
        B_u=Tx_inv @ model.B_u,
        C_y=model.C_y @ Tx,
        D_yu=model.D_yu,
        ts=model.ts,
        norm=model.norm,
    )


def compute_nonparametric(U: np.ndarray, Y: np.ndarray) -> NonparametricBLA:
    """Compute nonparametric BLA and variance estimates from input-output data.

    Parameters
    ----------
    U : np.ndarray, shape (F, nu, R, P)
        DFT input spectrum at the excited frequencies across realizations and
        periods.
    Y : np.ndarray, shape (F, ny, R, P)
        DFT output spectrum at the excited frequencies across realizations and
        periods.

    Returns
    -------
    `NonparametricBLA`
        Nonparametric BLA estimate with frequency response and variance
        estimates.

    """
    G = compute_frequency_response(U, Y)

    M, P = G.shape[3:5]

    # Compute noise variance
    G_P = G.mean(axis=4)  # shape (F, ny, nu, M)
    if P > 1:
        sqr_error = np.abs(G - G_P[..., None]) ** 2  # shape (F, ny, nu, M, P)
        tot_sqr_error = sqr_error.sum(axis=(3, 4))  # shape (F, ny, nu)
        var_noise = tot_sqr_error / (M * (P - 1))  # shape (F, ny, nu)
        var_noise = jnp.asarray(var_noise)
    else:
        var_noise = None

    # Compute total variance
    G_bla = G_P.mean(axis=3)  # shape (F, ny, nu)
    if M > 1:
        sqr_error = np.abs(G_P - G_bla[..., None]) ** 2  # shape (F, ny, nu, M)
        tot_sqr_error = sqr_error.sum(axis=3)  # shape (F, ny, nu)
        var_tot = tot_sqr_error / (M - 1)  # shape (F, ny, nu)
        var_tot = jnp.asarray(var_tot)
    else:
        var_tot = None

    G_bla = jnp.asarray(G_bla)

    return NonparametricBLA(G_bla, var_noise, var_tot)


def subspace_id(data: InputOutputData, nx: int, q: int) -> ModelBLA:
    """Parametrize a state-space model using the frequency-domain subspace method.

    Parameters
    ----------
    data : `InputOutputData`
        Estimation data.
    nx : int
        State dimension of the system to be identified.
    q : int
        Subspace dimensioning parameter, must be greater than `nx`.

    Returns
    -------
    `ModelBLA`
        Estimated state-space model in BLA form.

    Raises
    ------
    ValueError
        If `q `is not greater than `nx`.

    """
    header = " Frequency-domain subspace identification "
    print(f"{header:=^72}")
    if q <= nx:
        raise ValueError(
            f"Subspace dimension q={q} must be greater than state dimension nx={nx}."
        )

    freq = data.freq
    f_data = freq.f[freq.f_idx]
    fs = freq.fs
    ts = 1 / fs
    z = 2 * np.pi * f_data / fs

    G_bla = freq.G_bla
    F, ny, nu = G_bla.G.shape

    # Convert BLA to input-output form for FSID algorithm compatibility
    Y = np.transpose(G_bla.G, (0, 2, 1)).reshape(nu * F, ny)
    U = np.tile(np.eye(nu), (F, 1))
    zj = np.repeat(np.exp(z * 1j), nu)

    # Create weighting matrix (inverse of total variance)
    if G_bla.var_tot is not None:
        W_temp = 1 / G_bla.var_tot

        # The steps below are to make it compatible with fsid.gfdsid
        W_temp = np.transpose(np.sqrt(W_temp), (0, 2, 1)).reshape(nu * F, ny)
        W = np.zeros((nu * F, ny, ny))
        for k in range(nu * F):
            np.fill_diagonal(W[k], W_temp[k])
    else:
        W = np.empty(0)

    # Ensure that zj, Y, and U are numpy arrays
    zj = np.asarray(zj, dtype=np.complex128)
    Y = np.asarray(Y, dtype=np.complex128)
    U = np.asarray(U, dtype=np.complex128)

    # Perform frequency subspace identification
    fddata = (zj, Y, U)
    A, B_u, C_y, D_yu = fsid.gfdsid(fddata=fddata, n=nx, q=q, estTrans=False, w=W)[:4]

    model = ModelBLA(A, B_u, C_y, D_yu, ts, data.norm)

    _misc.evaluate_model_performance(model, data)

    return model


def optimize(
    model: ModelBLA,
    data: InputOutputData,
    *,
    solver: optx.AbstractLeastSquaresSolver | optx.AbstractMinimiser = SOLVER,
    max_iter: int = MAX_ITER,
) -> ModelBLA:
    """Refine the parameters of the BLA using frequency-response computations.

    Parameters
    ----------
    model : `ModelBLA`
        Initial BLA model to be optimized.
    data : `InputOutputData`
        Estimation data.
    solver : `optx.AbstractLeastSquaresSolver` or `optx.AbstractMinimiser`
        Any least-squares solver or general minimization solver from the
        Optimistix or Optax libraries.
    max_iter : int
        Maximum number of optimization iterations.

    Returns
    -------
    `ModelBLA`
        BLA model with optimized parameters.

    """
    header = " BLA optimization  "
    print(f"{header:=^72}")

    freq = data.freq
    G_bla = freq.G_bla
    f_data = freq.f[freq.f_idx]

    # Create weighting matrix (inverse of total variance)
    if G_bla.var_tot is not None:
        W = 1 / G_bla.var_tot
    else:
        W = jnp.ones_like(G_bla.G)

    model = _normalize_states(model, freq)
    theta0_dyn, theta_static = eqx.partition(model, eqx.is_inexact_array)

    args = (theta_static, jnp.asarray(G_bla.G), f_data, W)

    # Optimize the model parameters
    print("Starting iterative optimization...")
    solve_result = solve(theta0_dyn, solver, args, _compute_weighted_residual, max_iter)

    model = eqx.combine(solve_result.theta, theta_static)
    model = _normalize_states(model, freq)

    _misc.evaluate_model_performance(model, data, solve_result=solve_result)

    return model



"""Quick example on the Silverbox benchmark dataset using a neural network."""

import jax

import freq_statespace as fss


data = fss.load_and_preprocess_silverbox_data()  # 8192 x 6 samples

# Step 1: BLA estimation
nx = 2  # state dimension
q = nx + 1  # subspace dimensioning parameter
bla = fss.lin.subspace_id(data, nx, q)  # NRMSE 18.36%, non-iterative
bla = fss.lin.optimize(bla, data)  # NRMSE 13.17%, 6 iters, 1.67ms/iter

# Step 2: Nonlinear optimization
neural_net = fss.f_static.NeuralNetwork(
    nw=1, nz=1, num_layers=1, num_neurons_per_layer=10, activation=jax.nn.relu
)
nllfr = fss.nonlin.construct(bla, neural_net)
nllfr = fss.nonlin.optimize(nllfr, data)  # NRMSE 0.54%, 100 iters, 356ms/iter



"""General static nonlinear function mappings (mapping `z` to `w`)."""

from abc import abstractmethod
from collections.abc import Callable

import equinox as eqx
import jax
import jax.numpy as jnp

from freq_statespace import _misc
from freq_statespace._config import SEED
from freq_statespace.f_static._feature_maps import AbstractFeatureMap


class AbstractNonlinearFunction(eqx.Module, strict=True):
    """Abstract base class for nonlinear function mappings.

    Subclasses must provide the attributes `nw`, `nz` and `seed`,
    and must implement the methods `_evaluate` and `num_parameters`.
    """

    nw: eqx.AbstractVar[int]
    nz: eqx.AbstractVar[int]
    seed: eqx.AbstractVar[int]

    @abstractmethod
    def _evaluate(self, z: jnp.ndarray) -> jnp.ndarray:
        """From size (-1, `nz`) to size (-1, `nw`)."""
        pass

    @abstractmethod
    def num_parameters(self) -> int:
        """Return the total number of model parameters."""
        pass


class BasisFunctionModel(AbstractNonlinearFunction, strict=True):
    """Static nonlinear function based on an `AbstractFeatureMap`.

    Attributes
    ----------
    nw : int
        Number of output features (dimension of latent signal `w`).
    phi : `AbstractFeatureMap`
        Basis function that is linear in the parameters.
    seed : int
        Used for randomly initializing:
        1. Nonlinear coefficient matrix `beta`;
        2. The matrices `B_w`, `C_z`, `D_yw`, and `D_zu` (initialized
           externally, not by this class).

    """

    nw: int
    phi: AbstractFeatureMap
    seed: int = eqx.field(default=SEED, repr=False)

    # Post-init attributes
    nz: int = eqx.field(init=False)
    beta: jnp.array = eqx.field(init=False)
    _num_parameters: int = eqx.field(init=False, repr=False)

    def __post_init__(self) -> None:
        """Parametrize `beta` given random seed."""
        key = _misc.get_key(self.seed, "nonlin_funcs")
        self.beta = jax.random.uniform(
            key, shape=(self.phi.num_features(), self.nw), minval=-1, maxval=1
        )
        self.nz = self.phi.nz
        self._num_parameters = self.beta.size

    def _evaluate(self, z: jnp.ndarray) -> jnp.ndarray:
        return self.phi._compute_features(z) @ self.beta

    def num_parameters(self) -> int:
        """Return the size of `beta`."""
        return self._num_parameters


class NeuralNetwork(AbstractNonlinearFunction, strict=True):
    """Fully connected feedforward neural network.

    Attributes
    ----------
    nz : int
        Number of input features (dimension of latent signal `z`).
    nw : int
        Number of output features (dimension of latent signal `w`).
    num_layers : int
        Number of hidden layers.
    num_neurons_per_layer : int
        Number of neurons per hidden layer.
    activation : Callable, from `jax.nn`
        Activation function used in each hidden layer.
    seed : int
        Used for randomly initializing:
        1. Nonlinear coefficient matrix `beta`;
        2. The matrices `B_w`, `C_z`, `D_yw`, and `D_zu` (initialized
           externally, not by this class).
    bias : bool
        Whether to include bias terms in each layer.

    """

    nz: int = eqx.field(repr=False)
    nw: int = eqx.field(repr=False)
    num_layers: int = eqx.field(repr=False)
    num_neurons_per_layer: int = eqx.field(repr=False)
    activation: Callable = eqx.field(repr=False)
    seed: int = eqx.field(default=SEED, repr=False)
    bias: bool = eqx.field(default=True, repr=False)

    # Post-init attributes
    model: eqx.nn.MLP = eqx.field(init=False)
    _num_parameters: int = eqx.field(init=False, repr=False)

    def __post_init__(self) -> None:
        """Parametrize neural network given random seed."""
        key = _misc.get_key(self.seed, "nonlin_funcs")
        self.model = eqx.nn.MLP(
            in_size=self.nz,
            out_size=self.nw,
            width_size=self.num_neurons_per_layer,
            depth=self.num_layers,
            activation=self.activation,
            use_bias=self.bias,
            key=key,
        )
        self._num_parameters = sum(
            x.size
            for x in jax.tree_util.tree_leaves(self.model)
            if isinstance(x, jax.Array)
        )

    def _evaluate(self, z: jnp.ndarray) -> jnp.ndarray:
        return jax.vmap(self.model)(z)

    def num_parameters(self) -> int:
        """Return the number of neural network parameters."""
        return self._num_parameters



"""Quick example on the Silverbox benchmark dataset using inference and learning."""

import freq_statespace as fss


data = fss.load_and_preprocess_silverbox_data()  # 8192 x 6 samples

# Step 1: BLA estimation
nx = 2  # state dimension
q = nx + 1  # subspace dimensioning parameter
bla = fss.lin.subspace_id(data, nx, q)  # NRMSE 18.36%, non-iterative
bla = fss.lin.optimize(bla, data)  # NRMSE 13.17%, 6 iters, 1.97ms/iter

# Step 2: Inference and learning
phi = fss.f_static.basis.Polynomial(nz=1, degree=3)
nllfr = fss.nonlin.inference_and_learning(
    bla, data, phi=phi, nw=1, lambda_w=1e-2, fixed_point_iters=5
)  # NRMSE 1.11%, 42 iters, 13.2ms/iter

# Step 3: Nonlinear optimization
nllfr = fss.nonlin.optimize(nllfr, data)  # NRMSE 0.44%, 100 iters, 387ms/iter



"""Frequency response analysis for multi-input multi-output systems."""

import numpy as np


def compute_frequency_response(U: np.ndarray, Y: np.ndarray) -> np.ndarray:
    """Compute frequency response matrix G(k) = Y(k) * (U(k))^(-1).

    Parameters
    ----------
    U : np.ndarray, shape (F, nu, R, P)
        DFT input spectra at the excited frequencies across realizations and
        periods.
    Y : np.ndarray, shape (F, ny, R, P)
        DFT output spectra at the excited frequencies across realizations and
        periods.

    Returns
    -------
    G : np.ndarray, shape (F, ny, nu, M, P)
        Frequency Response Matrix:
        - F: number of frequency bins;
        - ny: number of outputs;
        - nu: number of inputs;
        - M: number of experiments (R // nu);
        - P: number of periods.

    """
    F, nu, R, P = U.shape
    ny = Y.shape[1]

    if R < nu:
        raise ValueError(
            "For multi-input systems, the number of realizations (R) must be "
            "at least equal to the number of inputs (nu) to compute the "
            "frequency response matrix."
        )

    M = R // nu
    if M * nu != R:
        print(
            "Suboptimal number of realizations: not all realizations are "
            "used to compute the frequency response matrix. Ideally, the "
            "number of realizations (R) should be an integer multiple of "
            "the number of inputs (nu)."
        )

    G = np.zeros((F, ny, nu, M, P), dtype=complex)

    for kf in range(F):
        for kr in range(M):
            for kp in range(P):
                start_idx = kr * nu
                end_idx = (kr + 1) * nu
                U_block = U[kf, :, start_idx:end_idx, kp]
                Y_block = Y[kf, :, start_idx:end_idx, kp]

                U_inv = np.linalg.solve(U_block, np.eye(nu))
                G[kf, :, :, kr, kp] = Y_block @ U_inv

    return G
